#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <SoftwareSerial.h>

#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 32
#define TEMP_PIN A0
#define OLED_DISPLAY_ADDRESS 0x3C  // Address 0x3D for 128x32

#define BLUETOOTH_RX_PIN 11
#define BLUETOOTH_TX_PIN 10

Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, -1);

SoftwareSerial BTSerial(BLUETOOTH_RX_PIN, BLUETOOTH_TX_PIN);

int interruptCount = 0;

void setup_debug_console() {
  Serial.begin(9600);
}

void setup_oled_display() {
  if (!display.begin(SSD1306_SWITCHCAPVCC, OLED_DISPLAY_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    for (;;)
      ;
  }
}

void setup_temperature_sensor() {
  ADMUX = 0;
  ADMUX |= (1 << REFS0) | (1 << REFS1);

  ADCSRA = 0;
  ADCSRA |= (7 << ADPS0);
  ADCSRA |= (1 << ADEN);
}

void setup_bluetooth() {
  BTSerial.begin(9600);
}

float read_temperature() {
  ADCSRA |= (1 << ADSC);
  while (!(ADCSRA & (1 << ADIF)))
    ;
  float reading = ADC;
  float temp = (reading - 500.0) / 10.0;  // 1 deg. per 10mV linear
  return temp;
}

void display_data(float temp, float pulse) {
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(WHITE);
  display.setCursor(0, 0);

  // Display static text
  display.print("Temperatura: ");
  display.print(temp, 1);
  display.println("C");
  display.display();
}

void send_data_over_bt(float temp, float pulse) {
  String temp_string = "Temperatura: " + String(temp) + " C";
  BTSerial.println(temp_string);
}

ISR(WDT_vect) {
  interruptCount += 1;
}

ISR(TIMER2_COMPA_vect) {
  interruptCount += 1;
}

void setup_interrupt_timer() {
  cli();
  // WDTCSR = (1 << WDCE);
  // WDTCSR |= (1 << WDP3);  //prescaler to 512K 4sec.
  // WDTCSR |= (1 << WDIE);
  // WDTCSR &= ~(0 << WDE);

  TCCR2A = 0;
  TCCR2B = 0;
  TCNT2 = 0;
  OCR2A = 255;
  TCCR2A |= (1 << WGM21);
  TCCR2B |= (7 << CS20);
  TIMSK2 |= (1 << OCIE2A);

  // Serial.print("WDTCSR: ");
  // Serial.println(WDTCSR & (1 << WDP3));
  sei();
}

void disable_unused_circuits() {
  PRR |= (1 << PRTIM0) | (1 << PRTIM1) | (1 << PRSPI);
}

void setup() {
  setup_debug_console();
  setup_temperature_sensor();
  setup_oled_display();
  setup_bluetooth();

  pinMode(13, OUTPUT);

  disable_unused_circuits();

  setup_interrupt_timer();
}

void loop() {
  // enter_power_save_mode();
  // wake_up_from_power_save_mode();

  if (interruptCount == 304) {
    float temp_value = read_temperature();
    display_data(temp_value, 0);
    send_data_over_bt(temp_value, 0);
    interruptCount = 0;
  }
}

void enter_power_save_mode() {
  SMCR |= (1 << SM0) | (1 << SM1) | (1 << SE);
  digitalWrite(13, LOW);
  __asm__ __volatile__ ( "sleep" "\n\t" :: );
}

void wake_up_from_power_save_mode() {
  SMCR &= ~(1 << SE);
  digitalWrite(13, HIGH);
}
